import { ComputedRef } from 'vue';
import { Ref } from 'vue';
import { UnwrapNestedRefs } from 'vue';

export declare type FieldOptions<T> = <K extends FormKey<T>>(key: K) => FormField<T, K>;

declare type FindNestedValue<T extends Record<any, any> | never, Keys extends string[]> = T extends never ? never : (Keys extends [infer E, ...infer R] ? FindNestedValue<T[E], R extends string[] ? R : []> : T);

export declare type FormErrors<T extends Record<string, any>> = {
    [K in FormKey<T>]?: string[];
};

declare type FormField<T, K> = {
    getError: ValidationErrors;
    hasError: boolean;
    isDirty: ComputedRef<boolean>;
    onChange: (newValue: PathValue<T, K>) => void;
    value: PathValue<T, K>;
};

export declare type FormKey<T extends Record<string, any>> = RecursiveKeyOf<T>;

export declare type FormOptions<T extends Record<string, any>> = {
    /**
     * Triggers when form is changed
     */
    onChange?: (state: State<T>) => Promise<void> | void;
    /**
     * Local method to treat errors.
     * It's preferable to use {@link setDefaultOnError}
     *
     * @expects - the method to return { [key]: [errors messages] }
     * @example
     * ```Typescript
     * const { ... } = useForm(
     *	 {
     * 		name: 'Rimuru'
     *	 },
     *   {
     * 		onErrors: ({ form, isValid, errors }) => {
     * 			/// This method changes the native `isValid`
     * 			return true
     * 		}
     * 	 }
     * )
     * ```
     */
    onErrors?: OnErrors;
    /**
     * When true only returns errors if the key was `touched` (@default false)
     * * Note: Local onlyOnTouch takes priority over global onlyOnTouch
     * @default true
     */
    onlyOnTouchDefault?: boolean;
    /**
     * Triggers when form is submitted
     */
    onSubmit?: (state: State<T>) => Promise<void> | void;
    /**
     * Method called every time a value is changed
     *
     * @example
     * ```Typescript
     * const { ... } = useForm(
     *	 {
     * 		name: 'Rimuru'
     *	 },
     *   {
     * 		onTouch: (key) => { }
     * 	 }
     * )
     * ```
     */
    onTouch?: (key: FormKey<T>, value: unknown, previousValue: unknown) => void;
    /**
     * Method to validate form.
     * Usually with some kind of validator.
     *
     * @expects - to throw the errors,
     * 	so either `onError` or default method for `onError` {@link setDefaultOnError} can catch then
     * @param form - form state
     * @example
     * ```Typescript
     * const { ... } = useForm(
     *	 {
     * 		name: 'Rimuru'
     *	 },
     *   {
     * 		validate: (form) => {
     * 			throw new Error()
     * 		}
     * 	 }
     * )
     * ```
     */
    validate?: (form: T, changedKeys: Array<FormKey<T>>) => void | Promise<void> | ValidationErrors | Promise<ValidationErrors>;
    /**
     * Validate form.
     * When `true` every change batch will validate the form
     * With `false` will only validate on method {@link FormActions#handleSubmit}
     * or if {@link FieldOptions#validate}/{@link ProduceNewStateOptions#validate} is set `true`.
     *
     * * Note: Local {@link FieldOptions#validate} takes priority over global {@link FormOptions#validateDefault}
     * @default true
     */
    validateDefault?: boolean;
};

declare type FormOptions_2<T extends Record<string, any>> = {
    /**
     * Indicates whether the form state should be persisted in memory.
     */
    persistFormKey?: symbol;
    /**
     * Indicates whether the form should be reset to its initial state after it is submitted.
     */
    resetIsDirtyOnSubmit?: boolean;
    /**
     * A function that validates the form data. This function is invoked whenever the form data
     *
     * @param form
     * @param changedKeys
     * @returns
     */
    validate?: (form: UnwrapNestedRefs<T>, changedKeys: Array<FormKey<T>>) => void | ValidationErrors | Promise<void | ValidationErrors>;
    /**
     * Indicates whether the form should be validated by default.
     */
    validateDefault?: boolean;
};

/**
 * Represents the state of a form in your application. It provides various methods and properties
 * to interact with and manage the form's state and data.
 *
 * @template T - The type representing the shape of the form data.
 */
declare type FormState<T extends Record<string, any>> = {
    /**
     * Updates the value of a specific field in the form.
     *
     * @param key - The key (field name) of the field to update.
     * @param value - The new value to set for the field.
     */
    changeValue: <K extends keyof T>(key: FormKey<T>, value: UnwrapNestedRefs<T[K]>) => void;
    /**
     * An object containing validation errors for the form fields. The structure of this object
     * matches the shape of the form data (T).
     */
    errors: UnwrapNestedRefs<FormErrors<T>>;
    /**
     * Options and configurations for the form fields.
     */
    field: FieldOptions<T>;
    /**
     * Representation of the current state of the form data. This is a reactive object that reflects
     * changes made to the form fields.
     */
    form: UnwrapNestedRefs<T>;
    /**
     * Get the validation errors for a specific field in the form.
     *
     * @param key - The key (field name) of the field to retrieve validation errors for.
     * @returns An array of validation errors for the specified field.
     */
    getErrors: (key: FormKey<T>) => ValidationErrors;
    /**
     * Get the isDirty of a specific field in the form.
     *
     * @param key - The key (field name) of the field to retrieve the isDirty for.
     * @returns The isDirty of the specified field.
     */
    getIsDirtyField: <K extends keyof T>(key: FormKey<T>) => UnwrapNestedRefs<T[K]>;
    /**
     * Get the current value of a specific field in the form.
     *
     * @param key - The key (field name) of the field to retrieve the value for.
     * @returns The current value of the specified field.
     */
    getValue: <K extends keyof T>(key: FormKey<T>) => UnwrapNestedRefs<T[K]>;
    /**
     * Handles form submission by invoking the provided callback function when the form is submitted.
     * Returns a function that can be used as an event handler for form submission events.
     *
     * @param onSubmit - The callback function to be invoked when the form is submitted.
     * @param onInvalid - The callback when the form is invalid returns an object containing validation errors for the form fields.
     * @returns An event handler function for form submission.
     */
    handleSubmit: (onSubmit: (form: UnwrapNestedRefs<T>) => void, onInvalid?: (errors: UnwrapNestedRefs<FormErrors<T>>) => void) => (e?: Event) => void;
    /**
     * Checks if a specific field in the form has validation errors.
     *
     * @param key - The key (field name) of the field to check for validation errors.
     * @returns `true` if the field has validation errors, otherwise `false`.
     */
    hasError: (key: FormKey<T>) => boolean;
    /**
     * Indicates whether the form has been restored from a persisted state.
     */
    hasRestoredFromPersisted: Ref<boolean>;
    /**
     * Indicates whether the form is dirty.
     */
    isDirty: Ref<boolean>;
    /**
     * Indicates whether the form is currently in a valid state (i.e., no validation errors).
     */
    isValid: boolean;
    /**
     * Attaches an event listener to a specific form field. When the field's value changes,
     * the provided callback function is invoked with the new value.
     *
     * @param key - The key (field name) of the field to watch for changes.
     * @param newValue - The new value of the watched field.
     */
    onChange: <K extends keyof T>(key: FormKey<T>) => (newValue: UnwrapNestedRefs<T[K]>) => void;
    /**
     * Resets the form to a specified state by replacing the current form data with the provided
     * partial data object.
     *
     * @param newForm - The partial data object to set as the new form state.
     */
    reset: (newForm: Partial<T>) => void;
    /**
     * Resets the state of isDirty.
     */
    resetIsDirty: () => void;
    /**
     * Sets a validation error for the form or a specific field in the form.
     *
     * @param error - The validation error to set. It can be a single error message or an object
     * containing field-specific error messages.
     */
    setError: (error: ValidationError_2 | ValidationWithErrors) => void;
    /**
     * An object containing the touched state of the form fields. This object tracks which fields
     */
    touches: Touches<T>;
    /**
     * Registers a callback function to watch for changes in the form's state. When any of the form
     * fields change their values, the provided callback is invoked with an array of keys
     * representing the changed fields.
     *
     * @param callback - The callback function to be invoked when form fields change.
     */
    watch: (callback: (changedKeys: Array<FormKey<T>>) => void) => void;
};

export declare type GetErrors<T extends Record<string, any>> = string[] & FormErrors<T>;

export declare type GetErrorsOptions<T extends Record<string, any>> = {
    /**
     * Includes the children errors on the array
     */
    includeChildsIntoArray?: boolean;
    /**
     * Includes `key` in children paths
     */
    includeKeyInChildErrors?: boolean;
    /**
     * When true only returns if the key was `touched` (@default false)
     */
    onlyOnTouch?: boolean;
    /**
     * Array containing other keys to also validate on touch
     */
    onlyOnTouchKeys?: Array<FormKey<T>>;
    /**
     * Includes children errors as object into array.
     *
     * Note: If `includeChildsIntoArray` is true `strict`
     * will by default be false
     */
    strict?: boolean;
};

export declare type HasErrorOptions<T extends Record<string, any>> = {
    /**
     * When true only returns if the key was `touched` (@default false)
     */
    onlyOnTouch?: boolean;
    /**
     * Array containing other keys to also validate on touch
     */
    onlyOnTouchKeys?: Array<FormKey<T>>;
    /**
     * Includes object/array children
     */
    strict?: boolean;
};

export declare type OnErrors = (errors: any | any[]) => ValidationErrors;

declare type PathValue<T extends Record<string, any> | never, Keys extends FormKey<T> | keyof T> = FindNestedValue<T, SplitStringIntoStringArray<Keys extends string ? Keys : ''>>;

declare type RecursiveKeyOf<TObj extends object, isFirstLevel extends boolean = true> = {
    [TKey in keyof TObj & (string | number)]: isFirstLevel extends true ? RecursiveKeyOfHandleValue<TObj[TKey], `${TKey}`> : RecursiveKeyOfHandleValue<TObj[TKey], RecursiveKeyOfAccess<TKey>>;
}[keyof TObj & (string | number)];

declare type RecursiveKeyOfAccess<TKey extends string | number> = `.${TKey}`;

declare type RecursiveKeyOfHandleValue<TValue, Text extends string> = TValue extends Array<infer E> ? `${Text}` | RecursiveKeyOfHandleValue<E, `${Text}[${number}]`> : TValue extends object ? Text | `${Text}${RecursiveKeyOf<TValue, false>}` : Text;

export declare type ResetMethod<T extends Record<string, any>> = (newFrom: Partial<T>, resetOptions?: ResetOptions | undefined) => Promise<void>;

export declare type ResetOptions = {
    /**
     * On reset `touches` will be cleared
     *
     * @default true
     */
    clearTouched?: boolean;
};

declare type SplitStringIntoStringArray<K extends string> = K extends `${infer K1}.${infer R1}` ? [
...(K1 extends '' ? [] : SplitStringIntoStringArray<K1>),
...(R1 extends '' ? [] : SplitStringIntoStringArray<R1>)
] : (K extends `${infer V}[${infer N}]${infer R2}` ? [
...(V extends '' ? [] : SplitStringIntoStringArray<V>),
N,
...(R2 extends '' ? [] : SplitStringIntoStringArray<R2>)
] : K extends `${infer V}[${infer N}]` ? [...(V extends '' ? [] : SplitStringIntoStringArray<V>), N] : [K]);

declare type State<T extends Record<string, any>> = {
    errors: FormErrors<T>;
    form: T;
    touches: Touches<T>;
};

export declare type Touches<T extends Record<string, any>> = {
    [K in FormKey<T>]?: boolean;
};

export declare const useForm: <T extends Record<string, any>>(defaultValues: T, options?: FormOptions_2<T>) => FormState<T>;

declare type ValidationError = {
    error: string;
    path: string;
};

declare type ValidationError_2 = {
    error: string;
    path: string | string[];
};

export declare type ValidationErrors = Array<ValidationError>;

declare type ValidationWithErrors = {
    errors: string[];
    path: string | string[];
};

export { }
